# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.specs import native as native_specs
from orquesta.tests.unit.specs.native import base as test_base


class WorkflowSpecTest(test_base.OrchestraWorkflowSpecTest):
    def test_exception_empty_definition(self):
        self.assertRaises(ValueError, native_specs.WorkflowSpec, None, {})
        self.assertRaises(ValueError, native_specs.WorkflowSpec, None, "")
        self.assertRaises(ValueError, native_specs.WorkflowSpec, None, None)

    def test_basic_spec_instantiation(self):
        wf_name = "sequential"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertIsNotNone(wf_spec)
        self.assertEqual(wf_spec.description, "A basic sequential workflow.")
        self.assertIsInstance(wf_spec.tasks, native_specs.TaskMappingSpec)
        self.assertEqual(len(wf_spec.tasks), 3)

        self.assertListEqual(sorted(list(wf_spec.tasks.keys())), ["task1", "task2", "task3"])

        # Verify model for task1.
        task1 = wf_spec.tasks["task1"]

        self.assertIsInstance(task1, native_specs.TaskSpec)
        self.assertEqual(task1.action, "core.echo")
        self.assertDictEqual(task1.input, {"message": "<% ctx().name %>"})

        task1_transition_seqs = getattr(task1, "next")

        self.assertIsInstance(task1_transition_seqs, native_specs.TaskTransitionSequenceSpec)

        self.assertIsInstance(task1_transition_seqs[0], native_specs.TaskTransitionSpec)

        self.assertEqual(getattr(task1_transition_seqs[0], "when"), "<% succeeded() %>")

        self.assertListEqual(
            getattr(task1_transition_seqs[0], "publish"), [{"greeting": "<% result() %>"}]
        )

        self.assertEqual(getattr(task1_transition_seqs[0], "do"), "task2")

        # Verify model for task2.
        task2 = wf_spec.tasks["task2"]

        self.assertIsInstance(task2, native_specs.TaskSpec)
        self.assertEqual(task2.action, "core.echo")
        self.assertDictEqual(task2.input, {"message": "All your base are belong to us!"})

        task2_transition_seqs = getattr(task2, "next")

        self.assertIsInstance(task2_transition_seqs, native_specs.TaskTransitionSequenceSpec)

        self.assertIsInstance(task2_transition_seqs[0], native_specs.TaskTransitionSpec)

        self.assertEqual(getattr(task2_transition_seqs[0], "when"), "<% succeeded() %>")

        self.assertListEqual(
            getattr(task2_transition_seqs[0], "publish"),
            [{"greeting": '<% ctx("greeting") %>, <% result() %>'}],
        )

        self.assertListEqual(getattr(task2_transition_seqs[0], "do"), ["task3"])

        # Verify model for task3.
        task3 = wf_spec.tasks["task3"]

        self.assertIsInstance(task3, native_specs.TaskSpec)
        self.assertEqual(task3.action, "core.echo")
        self.assertDictEqual(task3.input, {"message": "<% ctx('greeting') %>"})

        task3_transition_seqs = getattr(task3, "next")

        self.assertIsInstance(task3_transition_seqs, native_specs.TaskTransitionSequenceSpec)

        self.assertIsInstance(task3_transition_seqs[0], native_specs.TaskTransitionSpec)

        self.assertEqual(getattr(task3_transition_seqs[0], "when"), "<% succeeded() %>")

        self.assertListEqual(
            getattr(task3_transition_seqs[0], "publish"), [{"greeting": "<% result() %>"}]
        )

        self.assertEqual(getattr(task3_transition_seqs[0], "do"), "continue")

    def test_basic_spec_serialization(self):
        wf_name = "sequential"
        wf_spec_1 = self.get_wf_spec(wf_name)
        wf_spec_2 = native_specs.WorkflowSpec.deserialize(wf_spec_1.serialize())

        self.assertDictEqual(wf_spec_2.serialize(), wf_spec_1.serialize())

    def test_get_start_tasks(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertListEqual(wf_spec.tasks.get_start_tasks(), [("task1", None, None)])

    def test_is_join_task(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertFalse(wf_spec.tasks.is_join_task("task4"))
        self.assertTrue(wf_spec.tasks.is_join_task("task7"))

    def test_is_split_task(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertTrue(wf_spec.tasks.is_split_task("task4"))
        self.assertFalse(wf_spec.tasks.is_split_task("task7"))

    def test_not_in_cycle(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertFalse(wf_spec.tasks.in_cycle("task4"))

    def test_has_cycles(self):
        wf_name = "cycle"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertTrue(wf_spec.tasks.has_cycles())

    def test_in_cycle(self):
        wf_name = "cycle"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertFalse(wf_spec.tasks.in_cycle("prep"))
        self.assertTrue(wf_spec.tasks.in_cycle("task1"))
        self.assertTrue(wf_spec.tasks.in_cycle("task2"))
        self.assertTrue(wf_spec.tasks.in_cycle("task3"))

    def test_in_cycle_of_multiple(self):
        wf_name = "cycles"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertFalse(wf_spec.tasks.in_cycle("prep"))
        self.assertTrue(wf_spec.tasks.in_cycle("task1"))
        self.assertTrue(wf_spec.tasks.in_cycle("task2"))
        self.assertTrue(wf_spec.tasks.in_cycle("task3"))
        self.assertTrue(wf_spec.tasks.in_cycle("task4"))
        self.assertTrue(wf_spec.tasks.in_cycle("task5"))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import six

from orquesta import exceptions as exc
from orquesta.expressions import base as expr_base
from orquesta.specs.mistral.v2 import base as mistral_spec_base
from orquesta.specs.mistral.v2 import tasks as task_models
from orquesta.specs.mistral.v2 import types as mistral_spec_types
from orquesta.specs import types as spec_types
from orquesta.utils import dictionary as dict_util


LOG = logging.getLogger(__name__)


def instantiate(definition):
    definition.pop("version", None)

    if len(definition.keys()) > 1:
        raise ValueError("Workflow definition contains more than one workflow.")

    wf_name, wf_spec = list(definition.items())[0]

    return WorkflowSpec(wf_spec, name=wf_name)


def deserialize(data):
    return WorkflowSpec.deserialize(data)


class WorkflowSpec(mistral_spec_base.Spec):
    _schema = {
        "type": "object",
        "properties": {
            "type": mistral_spec_types.WORKFLOW_TYPE,
            "vars": spec_types.NONEMPTY_DICT,
            "input": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
            "output": spec_types.NONEMPTY_DICT,
            "output-on-error": spec_types.NONEMPTY_DICT,
            "task-defaults": task_models.TaskDefaultsSpec,
            "tasks": task_models.TaskMappingSpec,
        },
        "required": ["tasks"],
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["input", "vars", "tasks", "output"]

    _context_inputs = ["input", "vars"]

    def render_input(self, runtime_inputs, in_ctx=None):
        input_specs = getattr(self, "input") or []
        default_inputs = dict([list(i.items())[0] for i in input_specs if isinstance(i, dict)])
        merged_inputs = dict_util.merge_dicts(default_inputs, runtime_inputs, True)
        rendered_inputs = {}
        errors = []

        try:
            rendered_inputs = expr_base.evaluate(merged_inputs, {})
        except exc.ExpressionEvaluationException as e:
            errors.append(str(e))

        return rendered_inputs, errors

    def render_vars(self, in_ctx):
        vars_specs = getattr(self, "vars") or {}
        rendered_vars = {}
        errors = []

        try:
            rendered_vars = expr_base.evaluate(vars_specs, in_ctx)
        except exc.ExpressionEvaluationException as e:
            errors.append(str(e))

        return rendered_vars, errors

    def render_output(self, in_ctx):
        output_specs = getattr(self, "output") or {}
        rendered_outputs = {}
        errors = []

        try:
            rendered_outputs = {
                var_name: expr_base.evaluate(var_expr, in_ctx)
                for var_name, var_expr in six.iteritems(output_specs)
            }
        except exc.ExpressionEvaluationException as e:
            errors.append(str(e))

        return rendered_outputs, errors


class WorkbookSpec(mistral_spec_base.Spec):
    _schema = {
        "type": "object",
        "properties": {
            "workflows": {
                "type": "object",
                "minProperties": 1,
                "patternProperties": {r"^(?!version)\w+$": WorkflowSpec},
            }
        },
        "additionalProperties": False,
    }



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NONEMPTY_STRING = {"type": "string", "minLength": 1}

UNIQUE_STRING_LIST = {"type": "array", "items": NONEMPTY_STRING, "uniqueItems": True, "minItems": 1}

POSITIVE_INTEGER = {"type": "integer", "minimum": 0}

POSITIVE_NUMBER = {"type": "number", "minimum": 0.0}

VERSION = {"anyOf": [NONEMPTY_STRING, POSITIVE_INTEGER, POSITIVE_NUMBER]}

ANY = {
    "anyOf": [
        {"type": "array"},
        {"type": "boolean"},
        {"type": "integer"},
        {"type": "number"},
        {"type": "object"},
        {"type": "string"},
    ]
}

ANY_NULLABLE = {
    "anyOf": [
        {"type": "null"},
        {"type": "array"},
        {"type": "boolean"},
        {"type": "integer"},
        {"type": "number"},
        {"type": "object"},
        {"type": "string"},
    ]
}

NONEMPTY_DICT = {
    "type": "object",
    "minProperties": 1,
    "patternProperties": {r"^\w+$": ANY_NULLABLE},
}

ONE_KEY_DICT = {
    "type": "object",
    "minProperties": 1,
    "maxProperties": 1,
    "patternProperties": {r"^\w+$": ANY_NULLABLE},
}

STRING_OR_POSITIVE_INTEGER = {"oneOf": [NONEMPTY_STRING, POSITIVE_INTEGER]}

STRING_OR_POSITIVE_NUMBER = {"oneOf": [NONEMPTY_STRING, POSITIVE_NUMBER]}

STRING_OR_BOOLEAN = {"oneOf": [NONEMPTY_STRING, {"type": "boolean"}]}

STRING_OR_ONE_KEY_DICT = {"oneOf": [NONEMPTY_STRING, ONE_KEY_DICT]}

UNIQUE_STRING_OR_ONE_KEY_DICT_LIST = {
    "type": "array",
    "items": STRING_OR_ONE_KEY_DICT,
    "uniqueItems": True,
    "minItems": 1,
}

UNIQUE_ONE_KEY_DICT_LIST = {
    "type": "array",
    "items": ONE_KEY_DICT,
    "uniqueItems": True,
    "minItems": 1,
}



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import inspect
import json
import jsonschema
import logging
import re
import six
import yaml

from orquesta import exceptions as exc
from orquesta.expressions import base as expr_base
from orquesta.specs import types as spec_types
from orquesta.utils import expression as expr_util
from orquesta.utils import parameters as args_util
from orquesta.utils import schema as schema_util
from orquesta.utils import strings as str_util


LOG = logging.getLogger(__name__)


def isspec(value):
    return inspect.isclass(value) and issubclass(value, Spec)


class Spec(object):
    _catalog = None

    _version = "1.0"

    _schema = {"type": "object"}

    _meta_schema = {
        "type": "object",
        "properties": {
            "name": spec_types.NONEMPTY_STRING,
            "version": spec_types.VERSION,
            "description": spec_types.NONEMPTY_STRING,
            "tags": spec_types.UNIQUE_STRING_LIST,
        },
    }

    _schema_validator = None

    # Put the name of the spec properties in the order of validation.
    _context_evaluation_sequence = []

    # Put the name of the spec properties that are inputs for the context.
    _context_inputs = []

    # Override __getattr__ so we can dynamically map class attributes to spec properties.
    # Per documentation, __getattr__ is called by __getattribute__ on AttributeError. In
    # this case, the attribute does not physically exist on the class and so __getattr__
    # is called which it is overridden here to access the spec dict.
    def __getattr__(self, name):
        # Retrieve from spec if attribute is a meta schema property.
        if name in self._meta_schema.get("properties", {}):
            return self.spec.get(name)

        if name.replace("_", "-") in self._meta_schema.get("properties", {}):
            return self.spec.get(name.replace("_", "-"))

        # Retrieve from spec if attribute is a schema property.
        if name in self._schema.get("properties", {}):
            return self.spec.get(name)

        if name.replace("_", "-") in self._schema.get("properties", {}):
            return self.spec.get(name.replace("_", "-"))

        # Retrieve from spec if attribute match a regex pattern in the schema.
        for pattern in self._schema.get("patternProperties", {}).keys():
            if re.match(pattern, name):
                return self.spec.get(name)

        # Use default for all other attributes.
        return self.__getattribute__(name)

    def __init__(self, spec, name=None, member=False):
        # Update the schema to include schema parts from parent classes.
        self._schema = self.get_schema(includes=None, resolve_specs=False)
        self._meta_schema = self.get_meta_schema()

        if not spec:
            raise ValueError("The spec cannot be type of None.")

        self.spec = (
            yaml.safe_load(spec)
            if not isinstance(spec, dict) and not isinstance(spec, list)
            else spec
        )

        if not isinstance(self.spec, dict) and not isinstance(spec, list):
            raise ValueError("The spec is not type of json or yaml.")

        if name:
            self.name = name

        self.member = member

        schema = (
            self._schema
            if member
            else schema_util.merge_schema(self.get_meta_schema(), self._schema)
        )

        # Process attributes defined under properties in the schema.
        property_specs = {k: v for k, v in six.iteritems(schema.get("properties", {})) if isspec(v)}

        for name, spec_cls in six.iteritems(property_specs):
            if self.spec.get(name):
                setattr(self, name, spec_cls(self.spec.get(name), member=True))

        # Process pattern properties (regex) defined in the schema.
        regex_property_specs = {
            k: v for k, v in six.iteritems(schema.get("patternProperties", {})) if isspec(v)
        }

        for pattern, spec_cls in six.iteritems(regex_property_specs):
            for name, value in six.iteritems(self.spec):
                if re.match(pattern, name) and value:
                    setattr(self, name, spec_cls(value, member=True))

    def copy(self):
        return self.deserialize(self.serialize())

    def serialize(self):
        value = {"catalog": self.get_catalog(), "version": self.get_version(), "spec": self.spec}

        if hasattr(self, "name") and self.name:
            value["name"] = self.name

        if hasattr(self, "member") and self.member:
            value["member"] = self.member

        return value

    @classmethod
    def deserialize(cls, data):
        if data.get("catalog") != cls.get_catalog():
            raise ValueError('Serialized spec catalog does not match "%s".' % cls.get_catalog())

        if data.get("version") != cls.get_version():
            raise ValueError('Serialized spec version does not match "%s".' % cls.get_version())

        return cls(data["spec"], name=data.get("name"), member=data.get("member", False))

    @classmethod
    def get_catalog(cls):
        return cls._catalog

    @classmethod
    def get_version(cls):
        return cls._version

    @classmethod
    def get_schema_validator(cls):
        if not cls._schema_validator:
            cls._schema_validator = jsonschema.Draft4Validator(cls.get_schema())

        return cls._schema_validator

    @classmethod
    def get_meta_schema(cls):
        meta_schema = {}

        bases = [b for b in cls.__bases__ if issubclass(b, Spec)]

        for base_cls in bases:
            parent_meta_schema = base_cls.get_meta_schema()
            meta_schema = schema_util.merge_schema(meta_schema, parent_meta_schema)

        return schema_util.merge_schema(meta_schema, cls._meta_schema)

    @classmethod
    def get_schema(cls, includes=["meta"], resolve_specs=True):
        schema = {}

        bases = [b for b in cls.__bases__ if issubclass(b, Spec)]

        for base_cls in bases:
            parent_schema = base_cls.get_schema(includes=None)
            schema = schema_util.merge_schema(schema, parent_schema)

        schema = schema_util.merge_schema(schema, cls._schema)

        if includes and "meta" in includes:
            meta_schema = cls.get_meta_schema()
            schema = schema_util.merge_schema(schema, meta_schema)

        if not resolve_specs:
            return schema

        # Resolve the schema for children specs under properties.
        for k, v in six.iteritems(schema.get("properties", {})):
            if inspect.isclass(v) and issubclass(v, Spec):
                schema["properties"][k] = v.get_schema(includes=None)

        # Resolve the schema for children specs under patternProperties.
        for k, v in six.iteritems(schema.get("patternProperties", {})):
            if inspect.isclass(v) and issubclass(v, Spec):
                schema["patternProperties"][k] = v.get_schema(includes=None)

        # Resolve the schema for children specs under items.
        items_schema = schema.get("items", {})

        if inspect.isclass(items_schema) and issubclass(items_schema, Spec):
            schema["items"] = items_schema.get_schema(includes=None)
        elif isinstance(items_schema, dict):
            for k, v in six.iteritems(items_schema.get("properties", {})):
                if inspect.isclass(v) and issubclass(v, Spec):
                    schema_properties = schema["items"]["properties"]
                    schema_properties[k] = v.get_schema(includes=None)

        return schema

    def get_spec_path(self, prop_name, parent=None):
        return (parent.get("spec_path") + "." + prop_name).strip(".") if parent else prop_name

    def get_schema_path(self, prop_name, parent=None):
        return (
            (parent.get("schema_path") + "." + "properties." + prop_name).strip(".")
            if parent
            else "properties." + prop_name
        )

    def inspect(self, app_ctx=None, raise_exception=False):
        if app_ctx and not isinstance(app_ctx, dict):
            raise TypeError("Application context is not type of dict.")

        errors = {}
        app_ctx_metadata = None

        def sort_errors(e):
            return (e["schema_path"], e["spec_path"])

        syntax_errors = sorted(self.inspect_syntax(), key=sort_errors)

        if syntax_errors:
            errors["syntax"] = syntax_errors

        semantic_errors = sorted(self.inspect_semantics(), key=sort_errors)

        if semantic_errors:
            errors["semantics"] = semantic_errors

        expr_errors = sorted(self.inspect_expressions(), key=sort_errors)

        if expr_errors:
            errors["expressions"] = expr_errors

        if app_ctx:
            app_ctx_metadata = {"ctx": app_ctx.keys(), "spec_path": ".", "schema_path": "."}

        ctx_errors, _ = self.inspect_context(parent=app_ctx_metadata)
        ctx_errors = sorted(ctx_errors, key=sort_errors)

        if ctx_errors:
            errors["context"] = ctx_errors

        if errors and raise_exception:
            raise exc.WorkflowInspectionError(errors)

        return errors

    def inspect_syntax(self):
        result = []
        validator = self.get_schema_validator()

        for e in validator.iter_errors(self.spec):
            spec_path = ""
            schema_path = ""

            for s in list(e.absolute_path):
                spec_path += (
                    "[" + str(s) + "]" if isinstance(s, int) else "." + s if spec_path else s
                )

            for s in list(e.absolute_schema_path):
                schema_path += (
                    "[" + str(s) + "]" if isinstance(s, int) else "." + s if schema_path else s
                )

            entry = {
                "message": str_util.unescape(e.message),
                "spec_path": spec_path or None,
                "schema_path": schema_path or None,
            }

            result.append(entry)

        return result

    def inspect_semantics(self, parent=None):
        if parent and not parent.get("spec_path", None):
            raise ValueError("Parent context is missing spec path.")

        if parent and not parent.get("schema_path", None):
            raise ValueError("Parent context is missing schema path.")

        errors = []
        properties = {}
        schema = self.get_schema(includes=None)

        for prop_name, prop_type in six.iteritems(schema.get("properties", {})):
            properties[prop_name] = getattr(self, prop_name)

        for prop_name_regex, prop_type in six.iteritems(schema.get("patternProperties", {})):
            for prop_name in [key for key in self.keys() if re.findall(prop_name_regex, key)]:
                properties[prop_name] = getattr(self, prop_name)

        for prop_name, prop_value in six.iteritems(properties):
            spec_path = self.get_spec_path(prop_name, parent=parent)
            schema_path = self.get_schema_path(prop_name, parent=parent)
            prop_parent = {"spec_path": spec_path, "schema_path": schema_path}

            if isinstance(prop_value, SequenceSpec):
                errors.extend(prop_value.inspect_semantics(parent=prop_parent))

                for i in range(0, len(prop_value)):
                    item = prop_value[i]
                    item_spec_path = spec_path + "[" + str(i) + "]"
                    item_schema_path = schema_path + ".items"
                    item_parent = {"spec_path": item_spec_path, "schema_path": item_schema_path}
                    errors.extend(item.inspect_semantics(parent=item_parent))

                continue

            if isinstance(prop_value, MappingSpec):
                errors.extend(prop_value.inspect_semantics(parent=prop_parent))

                for k, v in six.iteritems(prop_value):
                    item_spec_path = spec_path + "." + k
                    item_schema_path = schema_path + ".patternProperties.^\\w+$"
                    item_parent = {"spec_path": item_spec_path, "schema_path": item_schema_path}
                    errors.extend(v.inspect_semantics(parent=item_parent))

                continue

            if isinstance(prop_value, Spec):
                errors.extend(prop_value.inspect_semantics(parent=prop_parent))
                continue

        return errors

    def inspect_expressions(self, parent=None):
        if parent and not parent.get("spec_path", None):
            raise ValueError("Parent context is missing spec path.")

        if parent and not parent.get("schema_path", None):
            raise ValueError("Parent context is missing schema path.")

        errors = []
        properties = {}
        schema = self.get_schema(includes=None)

        for prop_name, prop_type in six.iteritems(schema.get("properties", {})):
            properties[prop_name] = getattr(self, prop_name)

        for prop_name_regex, prop_type in six.iteritems(schema.get("patternProperties", {})):
            for prop_name in [key for key in self.keys() if re.findall(prop_name_regex, key)]:
                properties[prop_name] = getattr(self, prop_name)

        for prop_name, prop_value in six.iteritems(properties):
            spec_path = self.get_spec_path(prop_name, parent=parent)
            schema_path = self.get_schema_path(prop_name, parent=parent)

            if isinstance(prop_value, SequenceSpec):
                for i in range(0, len(prop_value)):
                    item = prop_value[i]
                    item_spec_path = spec_path + "[" + str(i) + "]"
                    item_schema_path = schema_path + ".items"
                    item_parent = {"spec_path": item_spec_path, "schema_path": item_schema_path}
                    errors.extend(item.inspect_expressions(parent=item_parent))

                continue

            if isinstance(prop_value, MappingSpec):
                for k, v in six.iteritems(prop_value):
                    item_spec_path = spec_path + "." + k
                    item_schema_path = schema_path + ".patternProperties.^\\w+$"
                    item_parent = {"spec_path": item_spec_path, "schema_path": item_schema_path}
                    errors.extend(v.inspect_expressions(parent=item_parent))

                continue

            if isinstance(prop_value, Spec):
                item_parent = {"spec_path": spec_path, "schema_path": schema_path}
                errors.extend(prop_value.inspect_expressions(parent=item_parent))
                continue

            result = expr_base.validate(prop_value).get("errors", [])

            for entry in result:
                entry["spec_path"] = spec_path
                entry["schema_path"] = schema_path

            errors += result

        return errors

    def inspect_context(self, parent=None):
        if parent and not parent.get("spec_path", None):
            raise ValueError("Parent context is missing spec path.")

        if parent and not parent.get("schema_path", None):
            raise ValueError("Parent context is missing schema path.")

        def decorate_ctx_var(variable, spec_path, schema_path):
            return {
                "type": variable[0],
                "expression": variable[1],
                "name": variable[2],
                "spec_path": spec_path,
                "schema_path": schema_path,
            }

        def decorate_ctx_var_error(var, msg):
            error = expr_util.format_error(
                var["type"], var["expression"], msg, var["spec_path"], var["schema_path"]
            )

            return error

        def get_ctx_inputs(prop_name, prop_value):
            ctx_inputs = []

            # By default, context inputs support only dictionary,
            # list of single item dictionaries, or string spec_types.
            if isinstance(prop_value, dict):
                ctx_inputs = list(prop_value.keys())
            elif isinstance(prop_value, list):
                for prop_value_item in prop_value:
                    if isinstance(prop_value_item, six.string_types):
                        ctx_inputs.append(prop_value_item)
                    elif isinstance(prop_value_item, dict) and len(prop_value_item) == 1:
                        ctx_inputs.extend(list(prop_value_item.keys()))
            elif isinstance(prop_value, six.string_types):
                ctx_inputs.append(prop_value)

            return ctx_inputs

        def inspect_ctx(prop_name, prop_value, spec_path, schema_path, rolling_ctx, errors):
            ctx_vars = []

            for var in expr_base.extract_vars(prop_value):
                ctx_vars.append(decorate_ctx_var(var, spec_path, schema_path))

            for ctx_var in ctx_vars:
                if ctx_var["name"].startswith("__"):
                    err_msg = (
                        'Variable "%s" that is prefixed with double underscores is considered '
                        "a private variable and cannot be referenced." % ctx_var["name"]
                    )
                    errors.append(decorate_ctx_var_error(ctx_var, err_msg))

                if ctx_var["name"] not in rolling_ctx:
                    err_msg = 'Variable "%s" is referenced before assignment.' % ctx_var["name"]
                    errors.append(decorate_ctx_var_error(ctx_var, err_msg))

            if prop_name in self._context_inputs:
                updated_ctx = get_ctx_inputs(prop_name, prop_value)
                rolling_ctx = list(set(rolling_ctx + updated_ctx))

            return rolling_ctx, errors

        errors = []
        parent_ctx = parent.get("ctx", []) if parent else []
        rolling_ctx = list(set(parent_ctx))

        for prop_name in self._context_evaluation_sequence:
            prop_value = getattr(self, prop_name)

            if not prop_value:
                continue

            spec_path = self.get_spec_path(prop_name, parent=parent)
            schema_path = self.get_schema_path(prop_name, parent=parent)

            # Pass the inspection downstream if value is a spec.
            if isinstance(prop_value, Spec):
                item_parent = {
                    "ctx": rolling_ctx,
                    "spec_path": spec_path,
                    "schema_path": schema_path,
                }

                result = prop_value.inspect_context(parent=item_parent)
                errors.extend(result[0])
                rolling_ctx = list(set(rolling_ctx + result[1]))

                continue

            # Parse inline parameters from value if value is a string.
            if isinstance(prop_value, six.string_types):
                inline_params = args_util.parse_inline_params(prop_value)

                if inline_params:
                    prop_value = inline_params

            # Preserve evaluation order if value is a list.
            if isinstance(prop_value, list):
                for i in range(0, len(prop_value)):
                    rolling_ctx, errors = inspect_ctx(
                        prop_name,
                        prop_value[i],
                        spec_path + "[" + str(i) + "]",
                        schema_path,
                        rolling_ctx,
                        errors,
                    )

                continue

            # Otherwise evaluate the value as is.
            rolling_ctx, errors = inspect_ctx(
                prop_name, prop_value, spec_path, schema_path, rolling_ctx, errors
            )

        return (sorted(errors, key=lambda x: x["spec_path"]), rolling_ctx)


class MappingSpec(Spec):
    def __setitem__(self, key, item):
        raise NotImplementedError()

    def __getitem__(self, key):
        if key not in self.keys():
            raise KeyError(key)

        return getattr(self, key)

    def __delitem__(self, key):
        raise NotImplementedError()

    def __repr__(self):
        return repr(self.spec)

    def __len__(self):
        return len(self.spec)

    def __contains__(self, item):
        return item in self.spec

    def __iter__(self):
        for key in self.keys():
            yield (key, getattr(self, key))

    def __unicode__(self):
        return str_util.unicode(repr(self.spec))

    def copy(self):
        name = self.name if "name" not in self.spec and hasattr(self, "name") else None

        return self.__class__(self.spec.copy(), name=name, member=self.member)

    def keys(self):
        return self.spec.keys()

    def values(self):
        return [getattr(self, key) for key in self.keys()]

    def items(self):
        return zip(self.keys(), self.values())

    def iteritems(self):
        for key in self.keys():
            yield (key, getattr(self, key))

    def clear(self):
        raise NotImplementedError()

    def pop(self, key, default=None):
        raise NotImplementedError()

    def update(self, *args, **kwargs):
        raise NotImplementedError()


class SequenceSpec(Spec, collections.MutableSequence):
    def __init__(self, spec, name=None, member=False):
        super(SequenceSpec, self).__init__(spec, name=name, member=member)

        schema = (
            self._schema
            if member
            else schema_util.merge_schema(self.get_meta_schema(), self._schema)
        )

        if schema.get("type") != "array":
            raise exc.SchemaDefinitionError("The schema for SequenceSpec must be type of array.")

        if not isspec(schema.get("items")):
            raise exc.SchemaDefinitionError("The item type for the array must be type of Spec.")

        if not isinstance(self.spec, list):
            raise ValueError("The spec is not type of list.")

        spec_cls = schema["items"]
        self._items = [spec_cls(item, member=True) for item in self.spec]

    def __len__(self):
        return len(self._items)

    def __getitem__(self, index):
        return self._items[index]

    def __delitem__(self, index):
        raise NotImplementedError()

    def __setitem__(self, index, item):
        raise NotImplementedError()

    def insert(self, index, item):
        raise NotImplementedError()

    def __str__(self):
        return json.dumps(self.spec, indent=4)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import statuses
from orquesta.tests.unit.conducting.native import base


class TaskTransitionWorkflowConductorTest(base.OrchestraWorkflowConductorTest):
    def test_on_error(self):
        wf_name = "error-handling"

        self.assert_spec_inspection(wf_name)

        # Mock task1 success
        expected_task_seq = ["task1", "task2"]

        mock_statuses = [statuses.SUCCEEDED, statuses.SUCCEEDED]  # task1  # task2

        expected_term_tasks = ["task2"]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_term_tasks=expected_term_tasks,
        )

        # Mock task1 error
        expected_task_seq = ["task1", "task3"]

        mock_statuses = [statuses.FAILED, statuses.SUCCEEDED]  # task1  # task3

        expected_term_tasks = ["task3"]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_term_tasks=expected_term_tasks,
        )

    def test_on_complete(self):
        wf_name = "task-on-complete"

        self.assert_spec_inspection(wf_name)

        # Mock task1 success
        expected_task_seq = ["task1", "task2", "task4"]

        expected_term_tasks = ["task2", "task4"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_term_tasks=expected_term_tasks,
        )

        # Mock task1 error
        expected_task_seq = ["task1", "task3", "task4"]

        mock_statuses = [
            statuses.FAILED,  # task1
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task4
        ]

        expected_term_tasks = ["task3", "task4"]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_term_tasks=expected_term_tasks,
        )

    def test_task_transitions_split(self):
        wf_name = "task-transitions-split"

        self.assert_spec_inspection(wf_name)

        # Mock task1 success
        expected_routes = [
            [],  # default from start
            ["task1__t0"],  # task1 -> task2 (when #1)
            ["task1__t2"],  # task1 -> task2 (when #3)
        ]

        expected_task_seq = [("task1", 0), ("task2", 1), ("task2", 2)]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1, 0
            statuses.SUCCEEDED,  # task2, 1 on complete
            statuses.SUCCEEDED,  # task2, 2 on success
        ]

        expected_term_tasks = [("task2", 1), ("task2", 2)]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

        # Mock task1 error
        expected_routes = [
            [],  # default from start
            ["task1__t0"],  # task1 -> task2 (when #1)
            ["task1__t1"],  # task1 -> task2 (when #2)
        ]

        expected_task_seq = [("task1", 0), ("task2", 1), ("task2", 2)]

        mock_statuses = [
            statuses.FAILED,  # task1, 0
            statuses.SUCCEEDED,  # task2, 1 on complete
            statuses.SUCCEEDED,  # task2, 2 on failure
        ]

        expected_term_tasks = [("task2", 1), ("task2", 2)]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import statuses
from orquesta.tests.unit.conducting.native import base


class SplitWorkflowConductorTest(base.OrchestraWorkflowConductorTest):
    def test_split(self):
        wf_name = "split"

        expected_routes = [
            [],  # default from start
            ["task2__t0"],  # task1 -> task2 -> task4
            ["task3__t0"],  # task1 -> task3 -> task4
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 0),
            ("task4", 1),
            ("task4", 2),
            ("task5", 1),
            ("task5", 2),
            ("task6", 1),
            ("task6", 2),
            ("task7", 1),
            ("task7", 2),
        ]

        expected_term_tasks = [("task7", 1), ("task7", 2)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

    def test_splits(self):
        wf_name = "splits"

        expected_routes = [
            [],  # default from start
            ["task1__t0"],  # task1 -> task8
            ["task2__t0"],  # task1 -> task2 -> task4
            ["task3__t0"],  # task1 -> task3 -> task4
            ["task2__t0", "task7__t0"],  # task1 -> task2 -> task4 -> task5/6 -> task7 -> task8
            ["task3__t0", "task7__t0"],  # task1 -> task3 -> task4 -> task5/6 -> task7 -> task8
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 0),
            ("task8", 1),
            ("task4", 2),
            ("task4", 3),
            ("task5", 2),
            ("task5", 3),
            ("task6", 2),
            ("task6", 3),
            ("task7", 2),
            ("task7", 3),
            ("task8", 4),
            ("task8", 5),
        ]

        expected_term_tasks = [("task8", 1), ("task8", 4), ("task8", 5)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

    def test_nested_splits(self):
        wf_name = "splits-nested"

        expected_routes = [
            [],  # default from start
            ["task2__t0"],  # task1 -> task2 -> task4
            ["task3__t0"],  # task1 -> task3 -> task4
            ["task2__t0", "task5__t0"],  # task1 -> task2 -> task4 -> task5 -> task7
            ["task3__t0", "task5__t0"],  # task1 -> task3 -> task4 -> task5 -> task7
            ["task2__t0", "task6__t0"],  # task1 -> task2 -> task4 -> task6 -> task7
            ["task3__t0", "task6__t0"],  # task1 -> task3 -> task4 -> task6 -> task7
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 0),
            ("task4", 1),
            ("task4", 2),
            ("task5", 1),
            ("task5", 2),
            ("task6", 1),
            ("task6", 2),
            ("task7", 3),
            ("task7", 4),
            ("task7", 5),
            ("task7", 6),
            ("task8", 3),
            ("task8", 4),
            ("task8", 5),
            ("task8", 6),
            ("task9", 3),
            ("task9", 4),
            ("task9", 5),
            ("task9", 6),
            ("task10", 3),
            ("task10", 4),
            ("task10", 5),
            ("task10", 6),
        ]

        expected_term_tasks = [("task10", 3), ("task10", 4), ("task10", 5), ("task10", 6)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

    def test_splits_mixed(self):
        wf_name = "splits-mixed"

        expected_routes = [
            [],
            ["task1__t0"],  # task1 -> task3
            ["task2__t0"],  # task2 -> task3
            ["task1__t0", "task3__t0"],  # task1 -> task3 -> task4
            ["task2__t0", "task3__t0"],  # task2 -> task3 -> task4
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 1),
            ("task3", 2),
            ("task4", 3),
            ("task4", 4),
            ("task5", 3),
            ("task5", 4),
        ]

        expected_term_tasks = [("task5", 3), ("task5", 4)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

    def test_splits_mixed_alt_branch(self):
        wf_name = "splits-mixed"

        expected_routes = [
            [],
            ["task1__t0"],  # task1 -> task3
            ["task2__t0"],  # task2 -> task3
            ["task1__t0", "task3__t0"],  # task1 -> task3 -> task4
            ["task2__t0", "task3__t1"],  # task2 -> task3 -> task4
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 1),
            ("task3", 2),
            ("task4", 3),
            ("task4", 4),
            ("task5", 3),
            ("task5", 4),
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1, 0
            statuses.SUCCEEDED,  # task2, 0
            statuses.SUCCEEDED,  # task3, 1
            statuses.FAILED,  # task3, 2
            statuses.SUCCEEDED,  # task4, 3
            statuses.SUCCEEDED,  # task4, 4
            statuses.SUCCEEDED,  # task5, 3
            statuses.SUCCEEDED,  # task5, 4
        ]

        expected_term_tasks = [("task5", 3), ("task5", 4)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

    def test_splits_multiple_transition(self):
        wf_name = "splits-transition"

        expected_routes = [
            [],
            ["task1__t0"],  # task1 -> task3
            ["task2__t0"],  # task2 -> task3
            ["task1__t0", "task3__t0"],  # task1 -> task3 -> task4 (when #1)
            ["task1__t0", "task3__t1"],  # task1 -> task3 -> task4 (when #2)
            ["task2__t0", "task3__t0"],  # task2 -> task3 -> task4 (when #1)
            ["task2__t0", "task3__t1"],  # task2 -> task3 -> task4 (when #2)
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 1),
            ("task3", 2),
            ("task4", 3),
            ("task4", 4),
            ("task4", 5),
            ("task4", 6),
            ("task5", 3),
            ("task5", 4),
            ("task5", 5),
            ("task5", 6),
            ("task6", 3),
            ("task6", 4),
            ("task6", 5),
            ("task6", 6),
            ("task7", 3),
            ("task7", 4),
            ("task7", 5),
            ("task7", 6),
        ]

        expected_term_tasks = [("task7", 3), ("task7", 4), ("task7", 5), ("task7", 6)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

    def test_very_many_splits(self):
        wf_name = "splits-very-many"

        expected_routes = [
            [],
            ["init__t1"],
            ["init__t0"],
            ["init__t0", "task2__t0"],
            ["init__t0", "task2__t0", "task7__t0"],
            ["init__t0", "task2__t0", "task7__t0", "task10__t0"],
            ["init__t0", "task2__t0", "task7__t0", "task10__t0", "task13__t0"],
            ["init__t0", "task2__t0", "task7__t0", "task10__t0", "task13__t0", "task15__t0"],
            [
                "init__t0",
                "task2__t0",
                "task7__t0",
                "task10__t0",
                "task13__t0",
                "task15__t0",
                "task17__t0",
            ],
        ]

        expected_task_seq = [
            ("init", 0),
            ("notify", 1),
            ("task2", 2),
            ("task5", 3),
            ("task7", 3),
            ("task9", 4),
            ("task10", 4),
            ("task11", 5),
            ("task12", 5),
            ("task13", 5),
            ("task14", 6),
            ("task15", 6),
            ("task17", 7),
            ("notify", 8),
        ]

        expected_term_tasks = [("notify", 1), ("notify", 8)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )

    def test_very_many_splits_alt(self):
        wf_name = "splits-very-many"

        inputs = {
            "fork1": True,
            "fork2": True,
            "fork3": True,
            "fork4": True,
            "fork5": True,
            "fork6": True,
            "fork7": True,
            "fork8": True,
        }

        mock_results = [
            None,  # init
            None,  # notify
            None,  # task1
            None,  # task2
            None,  # task3
            None,  # task4
            None,  # task5
            None,  # task6
            None,  # task8
            None,  # task9
            None,  # task11
            None,  # task12
            None,  # task13
            None,  # task14
            False,  # task16 (try #1)
            True,  # task16 (try #2)
            None,  # task17
            False,  # task18 (try #1)
            True,  # task18 (try #2)
            None,  # task19
            None,  # notify
        ]

        expected_routes = [
            [],
            ["init__t0"],
            ["task1__t0"],
            ["task1__t0", "task4__t0"],
            ["task1__t0", "task4__t0", "task6__t0"],
            ["task1__t0", "task4__t0", "task6__t0", "task8__t0"],
            ["task1__t0", "task4__t0", "task6__t0", "task8__t0", "task9__t0"],
            ["task1__t0", "task4__t0", "task6__t0", "task8__t0", "task9__t0", "task13__t0"],
            [
                "task1__t0",
                "task4__t0",
                "task6__t0",
                "task8__t0",
                "task9__t0",
                "task13__t0",
                "task16__t0",
            ],
            [
                "task1__t0",
                "task4__t0",
                "task6__t0",
                "task8__t0",
                "task9__t0",
                "task13__t0",
                "task16__t0",
                "task19__t0",
            ],
        ]

        expected_task_seq = [
            ("init", 0),
            ("notify", 1),
            ("task1", 0),
            ("task2", 2),
            ("task3", 2),
            ("task4", 2),
            ("task5", 3),
            ("task6", 3),
            ("task8", 4),
            ("task9", 5),
            ("task11", 6),
            ("task12", 6),
            ("task13", 6),
            ("task14", 7),
            ("task16", 7),
            ("task16", 7),
            ("task17", 8),
            ("task18", 8),
            ("task18", 8),
            ("task19", 8),
            ("notify", 9),
        ]

        expected_term_tasks = [("notify", 1), ("notify", 9)]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            inputs=inputs,
            mock_results=mock_results,
            expected_routes=expected_routes,
            expected_term_tasks=expected_term_tasks,
        )



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.composition.native import base


class CyclicWorkflowComposerTest(base.OrchestraWorkflowComposerTest):
    def test_cycle(self):
        wf_name = "cycle"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "prep"}, {"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [{"id": "task1", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "id": "task1",
                        "key": 0,
                        "ref": 0,
                        "criteria": ["<% succeeded() and ctx().count < 2 %>"],
                    }
                ],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "prep"}, {"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [{"id": "task1", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "id": "task1",
                        "key": 0,
                        "ref": 0,
                        "criteria": ["<% succeeded() and ctx().count < 2 %>"],
                    }
                ],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_cycles(self):
        wf_name = "cycles"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "prep"},
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4"},
                {"id": "task5"},
            ],
            "adjacency": [
                [{"id": "task1", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "ref": 0,
                        "criteria": ["<% succeeded() and not ctx().proceed %>"],
                    },
                    {
                        "id": "task5",
                        "key": 0,
                        "ref": 1,
                        "criteria": ["<% succeeded() and ctx().proceed %>"],
                    },
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "id": "task1",
                        "key": 0,
                        "ref": 0,
                        "criteria": ["<% succeeded() and ctx().count < 2 %>"],
                    }
                ],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "prep"},
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4"},
                {"id": "task5"},
            ],
            "adjacency": [
                [{"id": "task1", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "ref": 0,
                        "criteria": ["<% succeeded() and not ctx().proceed %>"],
                    },
                    {
                        "id": "task5",
                        "key": 0,
                        "ref": 1,
                        "criteria": ["<% succeeded() and ctx().proceed %>"],
                    },
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "id": "task1",
                        "key": 0,
                        "ref": 0,
                        "criteria": ["<% succeeded() and ctx().count < 2 %>"],
                    }
                ],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_rollback_retry(self):
        wf_name = "rollback-retry"

        expected_wf_ex_graph = {
            "directed": True,
            "graph": [],
            "nodes": [
                {"id": "init"},
                {"id": "create"},
                {"id": "rollback"},
                {"id": "check"},
                {"id": "delete"},
            ],
            "adjacency": [
                [
                    {"id": "create", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "check", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [],
                [{"id": "check", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "rollback", "key": 0, "ref": 1, "criteria": ["<% failed() %>"]},
                    {"id": "delete", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

REQUESTED = "requested"
SCHEDULED = "scheduled"
DELAYED = "delayed"
RUNNING = "running"
PENDING = "pending"
PAUSING = "pausing"
PAUSED = "paused"
RESUMING = "resuming"
SUCCEEDED = "succeeded"
FAILED = "failed"
EXPIRED = "timeout"
ABANDONED = "abandoned"
RETRYING = "retrying"
CANCELING = "canceling"
CANCELED = "canceled"
UNSET = "null"


ALL_STATUSES = [
    REQUESTED,
    SCHEDULED,
    DELAYED,
    RUNNING,
    PENDING,
    PAUSING,
    PAUSED,
    RESUMING,
    SUCCEEDED,
    FAILED,
    EXPIRED,
    ABANDONED,
    RETRYING,
    CANCELING,
    CANCELED,
    UNSET,
]

STARTING_STATUSES = [REQUESTED, SCHEDULED, DELAYED, RUNNING, PENDING]

RUNNING_STATUSES = [REQUESTED, SCHEDULED, DELAYED, RUNNING, RESUMING, RETRYING]

ACTIVE_STATUSES = [REQUESTED, SCHEDULED, DELAYED, RUNNING, RESUMING, PAUSING, CANCELING]

PAUSE_STATUSES = [PAUSING, PAUSED]

CANCEL_STATUSES = [CANCELING, CANCELED]

ABENDED_STATUSES = [FAILED, EXPIRED, ABANDONED]

COMPLETED_STATUSES = [SUCCEEDED, FAILED, EXPIRED, ABANDONED, CANCELED]


def is_valid(status):
    return status is None or status in ALL_STATUSES



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


LOG = logging.getLogger(__name__)


class PluginFactoryError(Exception):
    pass


class ExpressionGrammarException(Exception):
    pass


class ExpressionEvaluationException(Exception):
    pass


class VariableUndefinedError(Exception):
    def __init__(self, var):
        message = 'The variable "%s" is undefined.' % var
        super(VariableUndefinedError, self).__init__(message)


class VariableInaccessibleError(Exception):
    def __init__(self, var):
        message = 'The variable "%s" is for internal use and inaccessible.' % var
        super(VariableInaccessibleError, self).__init__(message)


class SchemaDefinitionError(Exception):
    pass


class SchemaIncompatibleError(Exception):
    pass


class InvalidTask(Exception):
    def __init__(self, task_id):
        message = 'Task "%s" does not exist.' % task_id
        super(InvalidTask, self).__init__(message)


class InvalidTaskTransition(Exception):
    def __init__(self, src, dest):
        message = 'Task transition from "%s" to "%s" does not exist.' % (src, dest)
        super(InvalidTaskTransition, self).__init__(message)


class AmbiguousTaskTransition(Exception):
    def __init__(self, src, dest):
        message = 'More than one task transitions found from "%s" to "%s".' % (src, dest)
        Exception.__init__(self, message)


class InvalidEventType(Exception):
    def __init__(self, type_name, event_name):
        message = 'Event type "%s" with event "%s" is not valid.' % (type_name, event_name)
        Exception.__init__(self, message)


class InvalidEvent(Exception):
    def __init__(self, value):
        message = 'Event "%s" is not valid.' % value
        super(InvalidEvent, self).__init__(message)


class InvalidStatus(Exception):
    def __init__(self, value):
        message = 'Status "%s" is not valid.' % value
        super(InvalidStatus, self).__init__(message)


class InvalidStatusTransition(Exception):
    def __init__(self, old, new):
        message = 'Status transition from "%s" to "%s" is invalid.' % (old, new)
        super(InvalidStatusTransition, self).__init__(message)


class InvalidTaskStatusTransition(Exception):
    def __init__(self, status, event):
        message = 'Unable to process event "%s" for task in "%s" status.' % (event, status)
        super(InvalidTaskStatusTransition, self).__init__(message)


class InvalidWorkflowStatusTransition(Exception):
    def __init__(self, status, event):
        message = 'Unable to process event "%s" for workflow in "%s" status.' % (event, status)
        super(InvalidWorkflowStatusTransition, self).__init__(message)


class InvalidTaskStateEntry(Exception):
    def __init__(self, task_id):
        message = 'Task "%s" is not staged or has not started yet.' % task_id
        super(InvalidTaskStateEntry, self).__init__(message)


class WorkflowInspectionError(Exception):
    def __init__(self, errors):
        message = "Workflow definition failed inspection."
        super(WorkflowInspectionError, self).__init__(message, errors)


class WorkflowContextError(Exception):
    pass


class WorkflowLogEntryError(Exception):
    pass


class WorkflowIsActiveAndNotRerunableError(Exception):
    def __init__(self):
        message = "Unable to rerun workflow because it is not in a completed state."
        super(WorkflowIsActiveAndNotRerunableError, self).__init__(message)


class InvalidTaskRerunRequest(Exception):
    def __init__(self, tasks):
        tasks_str = ""

        for task in tasks:
            tasks_str += ", " if tasks_str else ""
            tasks_str += "%s|%s" % (task.task_id, task.route)

        message = "Unable to rerun task|route(s) because it doesn't exist or isn't rerunnable: %s"
        super(InvalidTaskRerunRequest, self).__init__(message % tasks_str)


class UnreachableJoinError(Exception):
    def __init__(self, task_id, route):
        message = 'The join task|route "%s|%s" is partially satisfied but unreachable.'
        super(UnreachableJoinError, self).__init__(message % (task_id, route))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorContextTest(test_base.WorkflowConductorTest):
    def test_get_task_status_at_various_locations(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish: task_status_at_publish=<% task_status(task1) %>
                do: task2
          task2:
            action: core.echo
            input:
              message: <% task_status(task1) %>

        output:
          - task_status_at_publish: <% ctx().task_status_at_publish %>
          - task_status_at_output: <% task_status(task1) %>
        """

        expected_errors = []
        expected_output = {
            "task_status_at_publish": "succeeded",
            "task_status_at_output": "succeeded",
        }

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        # Run the workflow.
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assertListEqual(conductor.errors, expected_errors)

        # Complete task1.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Get next tasks and ensure task_status return expected result in task rendering.
        task2_ex_req = conductor.get_next_tasks()[0]
        self.assertEqual(task2_ex_req["id"], "task2")
        self.assertEqual(task2_ex_req["actions"][0]["action"], "core.echo")
        self.assertEqual(task2_ex_req["actions"][0]["input"]["message"], statuses.SUCCEEDED)

        # Complete task2.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_get_task_status_of_tasks_along_split(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.noop
            next:
              - do: task3
          task2:
            action: core.noop
            next:
              - do: task3
          task3:
            action: core.noop
            next:
              - publish:
                  - task1_status: <% task_status(task1) %>
                  - task2_status: <% task_status(task2) %>
                  - task3_status: <% task_status(task3) %>

        output:
          - task1_status: <% ctx(task1_status) %>
          - task2_status: <% ctx(task2_status) %>
          - task3_status: <% ctx(task3_status) %>
        """

        expected_errors = []
        expected_output = {
            "task1_status": "succeeded",
            "task2_status": "succeeded",
            "task3_status": "succeeded",
        }

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        # Run the workflow.
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assertListEqual(conductor.errors, expected_errors)

        # Complete tasks
        status_changes = [statuses.RUNNING, statuses.SUCCEEDED]
        self.forward_task_statuses(conductor, "task1", status_changes)
        self.forward_task_statuses(conductor, "task2", status_changes)
        self.forward_task_statuses(conductor, "task3", status_changes, route=1)
        self.forward_task_statuses(conductor, "task3", status_changes, route=2)

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_get_task_status_of_tasks_along_splits(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task2
              - when: <% failed() %>
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3
              - when: <% failed() %>
                do: task3
          task3:
            action: core.noop
            next:
              - publish:
                  - task1_status: <% task_status(task1) %>
                  - task2_status: <% task_status(task2) %>
                  - task3_status: <% task_status(task3) %>

        output:
          - task1_status: <% ctx(task1_status) %>
          - task2_status: <% ctx(task2_status) %>
          - task3_status: <% ctx(task3_status) %>
        """

        expected_errors = []
        expected_output = {
            "task1_status": "succeeded",
            "task2_status": "succeeded",
            "task3_status": "succeeded",
        }

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        # Run the workflow.
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assertListEqual(conductor.errors, expected_errors)

        # Complete tasks
        status_changes = [statuses.RUNNING, statuses.SUCCEEDED]
        self.forward_task_statuses(conductor, "task1", status_changes)
        self.forward_task_statuses(conductor, "task2", status_changes, route=1)
        self.forward_task_statuses(conductor, "task3", status_changes, route=2)

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorWithItemsTaskRenderingTest(test_base.WorkflowConductorTest):
    def test_bad_item_key(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: x in <% ctx(xs) %>
            action: core.echo message=<% item(y) %>
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression '<% item(y) %>'. "
                    'ExpressionEvaluationException: Item does not have key "y".'
                ),
                "task_id": "task1",
                "route": 0,
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        tasks = conductor.get_next_tasks()

        self.assertListEqual(tasks, [])
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_bad_item_type(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item(x) %>
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression '<% item(x) %>'. "
                    "ExpressionEvaluationException: Item is not type of collections.Mapping."
                ),
                "task_id": "task1",
                "route": 0,
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        tasks = conductor.get_next_tasks()

        self.assertListEqual(tasks, [])
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_bad_items_type(self):
        wf_def = """
        version: 1.0

        vars:
          - xs: fee fi fo fum

        tasks:
          task1:
            with: x in <% ctx(xs) %>
            action: core.echo message=<% item(y) %>
        """

        expected_errors = [
            {
                "type": "error",
                "message": 'TypeError: The value of "<% ctx(xs) %>" is not type of list.',
                "task_id": "task1",
                "route": 0,
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        tasks = conductor.get_next_tasks()

        self.assertListEqual(tasks, [])
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_start_task_rendering(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: x in <% ctx(xs) %>
            action: core.echo message=<% item(x) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        task_route = 0
        next_task_name = "task1"
        next_task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}
        next_task_spec = conductor.spec.tasks.get_task(next_task_name)

        next_task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        expected_task = self.format_task_item(
            next_task_name,
            task_route,
            next_task_ctx,
            next_task_spec,
            actions=next_task_action_specs,
            items_count=len(next_task_ctx["xs"]),
            items_concurrency=None,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

    def test_next_task_rendering(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            action: core.noop
            next:
              - do: task2
          task2:
            with: x in <% ctx(xs) %>
            action: core.echo message=<% item(x) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Process start task.
        task_route = 0
        next_task_name = "task1"
        next_task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}
        next_task_spec = conductor.spec.tasks.get_task(next_task_name)

        next_task_action_specs = [{"action": "core.noop", "input": None}]

        expected_task = self.format_task_item(
            next_task_name,
            task_route,
            next_task_ctx,
            next_task_spec,
            actions=next_task_action_specs,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        status_changes = [statuses.RUNNING, statuses.SUCCEEDED]
        self.forward_task_statuses(conductor, next_task_name, status_changes)

        # Process next task.
        next_task_name = "task2"
        next_task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}
        next_task_spec = conductor.spec.tasks.get_task(next_task_name)

        next_task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        expected_task = self.format_task_item(
            next_task_name,
            task_route,
            next_task_ctx,
            next_task_spec,
            actions=next_task_action_specs,
            items_count=len(next_task_ctx["xs"]),
            items_concurrency=None,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

    def test_basic_list_rendering(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        task_route = 0
        next_task_name = "task1"
        next_task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}
        next_task_spec = conductor.spec.tasks.get_task(next_task_name)

        next_task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        expected_task = self.format_task_item(
            next_task_name,
            task_route,
            next_task_ctx,
            next_task_spec,
            actions=next_task_action_specs,
            items_count=len(next_task_ctx["xs"]),
            items_concurrency=None,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

    def test_basic_list_rendering_var_w_in(self):
        wf_def = """
        version: 1.0

        vars:
          - domains:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(domains) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        task_route = 0
        next_task_name = "task1"
        next_task_ctx = {"domains": ["fee", "fi", "fo", "fum"]}
        next_task_spec = conductor.spec.tasks.get_task(next_task_name)

        next_task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        expected_task = self.format_task_item(
            next_task_name,
            task_route,
            next_task_ctx,
            next_task_spec,
            actions=next_task_action_specs,
            items_count=len(next_task_ctx["domains"]),
            items_concurrency=None,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

    def test_multiple_lists_rendering(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - foo
              - fu
              - marco
          - ys:
              - bar
              - bar
              - polo

        tasks:
          task1:
            with: x, y in <% zip(ctx(xs), ctx(ys)) %>
            action: core.echo message=<% item(x) + item(y) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        task_route = 0
        next_task_name = "task1"
        next_task_ctx = {"xs": ["foo", "fu", "marco"], "ys": ["bar", "bar", "polo"]}
        next_task_spec = conductor.spec.tasks.get_task(next_task_name)

        next_task_action_specs = [
            {"action": "core.echo", "input": {"message": "foobar"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fubar"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "marcopolo"}, "item_id": 2},
        ]

        expected_task = self.format_task_item(
            next_task_name,
            task_route,
            next_task_ctx,
            next_task_spec,
            actions=next_task_action_specs,
            items_count=len(next_task_ctx["xs"]),
            items_concurrency=None,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.expressions import jinja as jinja_expr
from orquesta.tests.unit import base as test_base
from orquesta.utils import plugin as plugin_util


class JinjaValidationTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "jinja"
        super(JinjaValidationTest, cls).setUpClass()

    def test_get_evaluator(self):
        self.assertEqual(
            plugin_util.get_module("orquesta.expressions.evaluators", self.language),
            jinja_expr.JinjaEvaluator,
        )

    def test_basic_validate(self):
        self.assertListEqual([], self.evaluator.validate("{{ 1 }}"))
        self.assertListEqual([], self.evaluator.validate("{{ abc }}"))
        self.assertListEqual([], self.evaluator.validate("{{ 1 + 2 }}"))
        self.assertListEqual([], self.evaluator.validate("{{ ctx().foo }}"))
        self.assertListEqual([], self.evaluator.validate('{{ ctx("foo") }}'))
        self.assertListEqual([], self.evaluator.validate('{{ ctx().a1 + ctx("a2") }}'))

    def test_parse_error(self):
        expr = "{{ {{ ctx().foo }} }}"
        errors = self.evaluator.validate(expr)

        self.assertEqual(2, len(errors))
        self.assertIn("expected token ':', got '}'", errors[0]["message"])
        self.assertIn("unexpected end of template", errors[1]["message"])

    def test_multiple_errors(self):
        expr = "{{ 1 +/ 2 }} and {{ * }}"
        errors = self.evaluator.validate(expr)

        self.assertEqual(3, len(errors))
        self.assertIn("unexpected", errors[0]["message"])
        self.assertIn("unexpected", errors[1]["message"])
        self.assertIn("unexpected", errors[2]["message"])

    def test_block_error(self):
        expr = "{% for i in ctx().x %}{{ i }}{% foobar %}"
        errors = self.evaluator.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("unknown tag", errors[0]["message"])

    def test_missing_braces_error(self):
        expr = "{% for i in ctx().x %}{{ i }}{{ foobar %}"
        errors = self.evaluator.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("unexpected '}'", errors[0]["message"])



# -*- coding: utf-8 -*-

# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base
from orquesta.utils import strings as str_util
import yaql.language.utils as yaql_utils


class WorkflowConductorDataFlowTest(test_base.WorkflowConductorTest):

    wf_def_yaql = """
    version: 1.0

    description: A basic sequential workflow.

    input:
      - a1
      - b1: <% ctx().a1 %>

    vars:
      - a2: <% ctx().b1 %>
      - b2: <% ctx().a2 %>

    output:
      - a5: <% ctx().b4 %>
      - b5: <% ctx().a5 %>

    tasks:
      task1:
        action: core.noop
        next:
          - when: <% succeeded() %>
            publish:
              - a3: <% ctx().b2 %>
              - b3: <% ctx().a3 %>
            do: task2
      task2:
        action: core.noop
        next:
          - when: <% succeeded() %>
            publish: a4=<% ctx().b3 %> b4=<% ctx().a4 %>
            do: task3
      task3:
        action: core.noop
    """

    wf_def_jinja = """
    version: 1.0

    description: A basic sequential workflow.

    input:
      - a1
      - b1: '{{ ctx("a1") }}'

    vars:
      - a2: '{{ ctx("b1") }}'
      - b2: '{{ ctx("a2") }}'

    output:
      - a5: '{{ ctx("b4") }}'
      - b5: '{{ ctx("a5") }}'

    tasks:
      task1:
        action: core.noop
        next:
          - when: '{{ succeeded() }}'
            publish:
              - a3: '{{ ctx("b2") }}'
              - b3: '{{ ctx("a3") }}'
            do: task2
      task2:
        action: core.noop
        next:
          - when: '{{ succeeded() }}'
            publish: a4='{{ ctx("b3") }}' b4='{{ ctx("a4") }}'
            do: task3
      task3:
        action: core.noop
    """

    def _prep_conductor(self, wf_def, context=None, inputs=None, status=None):
        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        kwargs = {
            "context": context if context is not None else None,
            "inputs": inputs if inputs is not None else None,
        }

        conductor = conducting.WorkflowConductor(spec, **kwargs)

        if status:
            conductor.request_workflow_status(status)

        return conductor

    def _get_combined_value(self, callstack_depth=0):
        # This returns dict typed value all Python built-in type values
        # which orquesta spec could accept.
        if callstack_depth < 2:
            return {
                "null": None,
                "integer_positive": 123,
                "integer_negative": -123,
                "number_positive": 99.99,
                "number_negative": -99.99,
                "string": "xyz",
                "boolean_true": True,
                "boolean_false": False,
                "array": list(self._get_combined_value(callstack_depth + 1).values()),
                "object": self._get_combined_value(callstack_depth + 1),
            }
        else:
            return {}

    def _assert_data_flow(self, inputs, expected_output):
        # This assert method checks input value would be handled and published
        # as an expected type and value with both YAQL and Jinja expressions.
        for wf_def in [self.wf_def_jinja, self.wf_def_yaql]:
            conductor = self._prep_conductor(wf_def, inputs=inputs, status=statuses.RUNNING)

            for i in range(1, len(conductor.spec.tasks) + 1):
                task_name = "task" + str(i)
                forward_statuses = [statuses.RUNNING, statuses.SUCCEEDED]
                self.forward_task_statuses(conductor, task_name, forward_statuses)

            # Render workflow output and checkout workflow status and output.
            conductor.render_workflow_output()
            self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
            self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def assert_data_flow(self, input_value):
        inputs = {"a1": input_value}
        expected_output = {"a5": inputs["a1"], "b5": inputs["a1"]}

        self._assert_data_flow(inputs, expected_output)

    def assert_unicode_data_flow(self, input_value):
        inputs = {
            u"a1": (
                str_util.unicode(input_value, encoding_type="utf-8", force=True)
                if six.PY2
                else input_value
            )
        }

        expected_output = {u"a5": inputs["a1"], u"b5": inputs["a1"]}

        self._assert_data_flow(inputs, expected_output)

    def test_data_flow_string(self):
        self.assert_data_flow("xyz")

    def test_data_flow_integer(self):
        self.assert_data_flow(123)
        self.assert_data_flow(-123)

    def test_data_flow_float(self):
        self.assert_data_flow(99.99)
        self.assert_data_flow(-99.99)

    def test_data_flow_boolean(self):
        self.assert_data_flow(True)
        self.assert_data_flow(False)

    def test_data_flow_none(self):
        self.assert_data_flow(None)

    def test_data_flow_dict(self):
        mapping_typed_data = self._get_combined_value()

        self.assertIsInstance(mapping_typed_data, yaql_utils.MappingType)
        self.assert_data_flow(mapping_typed_data)

    def test_data_flow_list(self):
        sequence_typed_data = list(self._get_combined_value().values())

        self.assertIsInstance(sequence_typed_data, yaql_utils.SequenceType)
        self.assert_data_flow(sequence_typed_data)

    def test_data_flow_unicode(self):
        self.assert_unicode_data_flow("光合作用")



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.expressions import jinja as jinja_expr
from orquesta.tests.unit import base as test_base
from orquesta.utils import plugin as plugin_util


class JinjaEvaluationTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "jinja"
        super(JinjaEvaluationTest, cls).setUpClass()

    def test_get_evaluator(self):
        e = plugin_util.get_module("orquesta.expressions.evaluators", self.language)

        self.assertEqual(e, jinja_expr.JinjaEvaluator)
        self.assertIn("ctx", e._custom_functions.keys())
        self.assertIn("task_status", e._custom_functions.keys())

    def test_block_eval(self):
        expr = "{% for i in ctx().x %}{{ i }}{% endfor %}"

        data = {"x": ["a", "b", "c"]}

        self.assertEqual("abc", self.evaluator.evaluate(expr, data))

    def test_block_eval_complex_data(self):
        expr = "{% for i in ctx().x %}{{ i.k }}{{ ctx().z }}{% endfor %}"

        data = {"x": [{"k": "a"}, {"k": "b"}, {"k": "c"}], "z": "->"}

        self.assertEqual("a->b->c->", self.evaluator.evaluate(expr, data))

    def test_block_eval_undefined(self):
        expr = "{% for i in ctx().x %}{{ ctx().y }}{% endfor %}"

        data = {"x": ["a", "b", "c"]}

        self.assertRaises(jinja_expr.JinjaEvaluationException, self.evaluator.evaluate, expr, data)

    def test_block_eval_recursive(self):
        expr = "{% for i in ctx().x %}{{ i }}{% endfor %}"

        data = {
            "x": [
                "{{ ctx().a }}",
                "{{ ctx().b }}",
                "{{ ctx().c }}" "{% for j in ctx().y %}{{ j }}{% endfor %}",
            ],
            "a": "a",
            "b": "b",
            "c": "c",
            "y": ["d", "e", "f"],
        }

        self.assertEqual("abcdef", self.evaluator.evaluate(expr, data))

    def test_multi_block_eval(self):
        expr = (
            "{% for i in ctx().x %}{{ i }}{% endfor %}" "{% for i in ctx().y %}{{ i }}{% endfor %}"
        )

        data = {"x": ["a", "b", "c"], "y": ["d", "e", "f"]}

        self.assertEqual("abcdef", self.evaluator.evaluate(expr, data))

    def test_raw_block(self):
        expr = "{% raw %}{{ ctx().foo }}{% endraw %}"

        data = {"foo": "bar"}

        self.assertEqual("{{ ctx().foo }}", self.evaluator.evaluate(expr, data))

    def test_multi_raw_blocks(self):
        expr = (
            "{% raw %}{{ ctx().foo }}{% endraw %} and "
            "{% raw %}{{ ctx().bar }}{% endraw %} and "
            "{% raw %}foobar{% endraw %}"
        )

        data = {"foo": "bar"}

        self.assertEqual(
            "{{ ctx().foo }} and {{ ctx().bar }} and foobar", self.evaluator.evaluate(expr, data)
        )

    def test_mix_block_and_expr_eval(self):
        expr = (
            "{{ ctx().a }}{% for i in ctx().x %}{{ i }}{% endfor %}{{ ctx().d }} and "
            "{% raw %}{{ 1234 }}{% endraw %}"
        )

        data = {"x": ["b", "c"], "a": "a", "d": "d"}

        self.assertEqual("abcd and {{ 1234 }}", self.evaluator.evaluate(expr, data))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import six

from orquesta import exceptions as exc


def json_(s):
    if isinstance(s, dict):
        return s

    if not isinstance(s, six.string_types):
        raise TypeError("Given object is not typeof string.")

    return json.loads(s)


def zip_(*args, **kwargs):
    args = [list() if arg is None else arg for arg in args]

    if len(args) == 1:
        return args[0]

    pad_with = kwargs.pop("pad", None)

    return list(six.moves.zip_longest(*args, fillvalue=pad_with))


def ctx_(context, key=None):
    if key:
        if key not in context["__vars"]:
            raise exc.VariableUndefinedError(key)

        if key in context["__vars"] and key.startswith("__"):
            raise exc.VariableInaccessibleError(key)

        return context["__vars"][key]
    else:
        return {k: v for k, v in six.iteritems(context["__vars"]) if not k.startswith("__")}



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hacking import core


CODE = "O102"

REQS = {
    "orquesta.composers.base": "comp_base",
    "orquesta.conducting": None,
    "orquesta.constants": None,
    "orquesta.events": None,
    "orquesta.graphing": None,
    "orquesta.exceptions": "exc",
    "orquesta.expressions.base": "expr_base",
    "orquesta.expressions.functions.base": "func_base",
    "orquesta.expressions.jinja": "jinja_expr",
    "orquesta.expressions.yql": "yaql_expr",
    "orquesta.machines": None,
    "orquesta.specs.base": "spec_base",
    "orquesta.specs.loader": "spec_loader",
    "orquesta.specs.mistral": "mistral_specs",
    "orquesta.specs.mistral.v2": "mistral_v2_specs",
    "orquesta.specs.mock": "mock_specs",
    "orquesta.specs.mock.models": "mock_models",
    "orquesta.specs.native": "native_specs",
    "orquesta.specs.native.v1": "native_v1_specs",
    "orquesta.specs.native.v1.models": "native_v1_models",
    "orquesta.specs.types": "spec_types",
    "orquesta.statuses": None,
    "orquesta.tests.fixtures.loader": "fixture_loader",
    "orquesta.tests.unit.base": "test_base",
    "orquesta.tests.unit.specs.base": "test_specs",
    "orquesta.utils.context": "ctx_util",
    "orquesta.utils.date": "date_util",
    "orquesta.utils.dictionary": "dict_util",
    "orquesta.utils.expression": "expr_util",
    "orquesta.utils.jsonify": "json_util",
    "orquesta.utils.parameters": "args_util",
    "orquesta.utils.plugin": "plugin_util",
    "orquesta.utils.schema": "schema_util",
    "orquesta.utils.specs": "spec_util",
    "orquesta.utils.strings": "str_util",
}


def get_alias(logical_line):

    parts = logical_line.split()

    if (
        len(parts) > 3
        and parts[0] == "from"
        and parts[1] != "__future__"
        and not core.is_import_exception(parts[1])
    ):

        # from path.to.module import module
        if len(parts) == 4:
            return ".".join([parts[1], parts[3]]), None

        # from path.to.module import module as alias
        if len(parts) > 5 and parts[4] == "as":
            return ".".join([parts[1], parts[3]]), parts[5]

    return None


@core.flake8ext
def check_alias_naming(logical_line, physical_line, filename, noqa):
    """Check import alias.

    To ensure code consistency, check if a module must use a specific alias.

    Examples:
    """

    if noqa:
        return

    alias = get_alias(logical_line)

    # Dictate import from using alias.
    if alias and alias[0] in REQS and alias[1] and not REQS[alias[0]]:
        message = "cannot use alias for import '%s'"
        yield 0, "%s %s" % (CODE, message % logical_line)
    # Dictate import to use alias.
    elif alias and alias[0] in REQS and not alias[1] and REQS[alias[0]]:
        message = "import '%s' must use alias '%s'"
        yield 0, "%s %s" % (CODE, message % (logical_line, REQS[alias[0]]))
    # Dictate import to use specific alias.
    elif alias and alias[0] in REQS and alias[1] and REQS[alias[0]] and alias[1] != REQS[alias[0]]:
        message = "import '%s' must use alias '%s'"
        yield 0, "%s %s" % (CODE, message % (logical_line, REQS[alias[0]]))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import copy
import six
from six.moves import queue
import unittest

from orquesta import conducting
from orquesta import events
from orquesta.expressions import base as expr_base
from orquesta.specs import loader as spec_loader
from orquesta import statuses
from orquesta.tests.fixtures import loader as fixture_loader
from orquesta.utils import plugin as plugin_util
from orquesta.utils import specs as spec_util


@six.add_metaclass(abc.ABCMeta)
class ExpressionEvaluatorTest(unittest.TestCase):
    language = None
    evaluator = None

    @classmethod
    def setUpClass(cls):
        cls.evaluator = plugin_util.get_module("orquesta.expressions.evaluators", cls.language)


@six.add_metaclass(abc.ABCMeta)
class ExpressionFacadeEvaluatorTest(unittest.TestCase):
    def validate(self, expr):
        return expr_base.validate(expr).get("errors", [])


@six.add_metaclass(abc.ABCMeta)
class WorkflowGraphTest(unittest.TestCase):
    def _zip_wf_graph_meta(self, wf_graph_json):
        wf_graph_json["adjacency"] = [
            sorted(link, key=lambda x: x["id"]) if link else link
            for link in wf_graph_json["adjacency"]
        ]

        wf_graph_meta = sorted(
            zip(wf_graph_json["nodes"], wf_graph_json["adjacency"]), key=lambda x: x[0]["id"]
        )

        return wf_graph_meta

    def assert_graph_equal(self, wf_graph, expected_wf_graph):
        wf_graph_json = wf_graph.serialize()
        wf_graph_meta = self._zip_wf_graph_meta(wf_graph_json)
        expected_wf_graph_meta = self._zip_wf_graph_meta(expected_wf_graph)

        self.assertListEqual(wf_graph_meta, expected_wf_graph_meta)

        wf_graph_attrs = sorted(wf_graph_json["graph"], key=lambda x: x[0])
        expected_wf_graph_attrs = sorted(expected_wf_graph["graph"], key=lambda x: x[0])

        self.assertListEqual(wf_graph_attrs, expected_wf_graph_attrs)


class WorkflowSpecTest(unittest.TestCase):
    spec_module_name = "mock"

    def __init__(self, *args, **kwargs):
        super(WorkflowSpecTest, self).__init__(*args, **kwargs)
        self.maxDiff = None

    def get_fixture_path(self, wf_name):
        return self.spec_module_name + "/" + wf_name + ".yaml"

    def get_wf_def(self, wf_name, raw=False):
        return fixture_loader.get_fixture_content(
            self.get_fixture_path(wf_name), "workflows", raw=raw
        )

    def get_wf_spec(self, wf_name):
        wf_def = self.get_wf_def(wf_name)
        wf_spec = spec_util.instantiate(self.spec_module_name, wf_def)
        return wf_spec

    def instantiate(self, wf_def):
        return spec_util.instantiate(self.spec_module_name, wf_def)

    def assert_spec_inspection(self, wf_name, errors=None):
        wf_spec = self.get_wf_spec(wf_name)

        self.assertDictEqual(wf_spec.inspect(), errors or {})


@six.add_metaclass(abc.ABCMeta)
class WorkflowComposerTest(WorkflowGraphTest, WorkflowSpecTest):
    composer = None

    @classmethod
    def setUpClass(cls):
        WorkflowGraphTest.setUpClass()
        WorkflowSpecTest.setUpClass()

        cls.composer = plugin_util.get_module("orquesta.composers", cls.spec_module_name)
        cls.spec_module = spec_loader.get_spec_module(cls.spec_module_name)
        cls.wf_spec_type = cls.spec_module.WorkflowSpec

    def compose_wf_graph(self, wf_name):
        wf_def = self.get_wf_def(wf_name)
        wf_spec = self.spec_module.instantiate(wf_def)

        return self.composer._compose_wf_graph(wf_spec)

    def assert_compose_to_wf_graph(self, wf_name, expected_wf_graph):
        wf_graph = self.compose_wf_graph(wf_name)

        self.assert_graph_equal(wf_graph, expected_wf_graph)

    def compose_wf_ex_graph(self, wf_name):
        wf_def = self.get_wf_def(wf_name)
        wf_spec = self.spec_module.instantiate(wf_def)

        return self.composer.compose(wf_spec)

    def assert_compose_to_wf_ex_graph(self, wf_name, expected_wf_ex_graph):
        wf_ex_graph = self.compose_wf_ex_graph(wf_name)

        self.assert_graph_equal(wf_ex_graph, expected_wf_ex_graph)


@six.add_metaclass(abc.ABCMeta)
class WorkflowConductorTest(WorkflowComposerTest):
    def format_task_item(
        self,
        task_id,
        route,
        ctx,
        spec,
        actions=None,
        delay=None,
        items_count=None,
        items_concurrency=None,
    ):

        if not actions and items_count is None:
            actions = [{"action": spec.action, "input": spec.input}]

        task = {"id": task_id, "route": route, "ctx": ctx, "spec": spec, "actions": actions or []}

        if delay:
            task["delay"] = delay

        if items_count is not None:
            task["items_count"] = items_count
            task["concurrency"] = items_concurrency

        return task

    def forward_task_statuses(self, conductor, task_id, status_changes, route=0, result=None):
        for status in status_changes:
            ac_ex_event = events.ActionExecutionEvent(status)

            if result is not None and status in statuses.COMPLETED_STATUSES:
                ac_ex_event.result = result

            conductor.update_task_state(task_id, route, ac_ex_event)

    def forward_task_item_statuses(
        self,
        conductor,
        task_id,
        item_id,
        status_changes,
        route=0,
        result=None,
        accumulated_result=None,
    ):
        for status in status_changes:
            ac_ex_event = events.TaskItemActionExecutionEvent(item_id, status)

            if accumulated_result is not None and status in statuses.COMPLETED_STATUSES:
                ac_ex_event.accumulated_result = accumulated_result

            if result is not None and status in statuses.COMPLETED_STATUSES:
                ac_ex_event.result = result

            conductor.update_task_state(task_id, route, ac_ex_event)

    # The conductor.get_next_tasks make copies of the task specs and render expressions
    # in the task action and task input. So comparing the task specs will not match. In
    # order to match in unit tests. This method is used to serialize the task specs and
    # compare the lists.
    def assert_task_list(self, conductor, actual, expected):
        actual_copy = copy.deepcopy(actual)
        expected_copy = copy.deepcopy(expected)

        for task in actual_copy:
            for staged_task in task["ctx"]["__state"]["staged"]:
                if "items" in staged_task:
                    del staged_task["items"]

            task["spec"] = task["spec"].serialize()

        for task in expected_copy:
            task["ctx"]["__current_task"] = {"id": task["id"], "route": task["route"]}
            task["ctx"]["__state"] = conductor.workflow_state.serialize()

            for staged_task in task["ctx"]["__state"]["staged"]:
                if "items" in staged_task:
                    del staged_task["items"]

            task["spec"] = task["spec"].serialize()

        self.assertListEqual(actual_copy, expected_copy)

    def assert_next_task(self, conductor, task_id=None, ctx=None, route=0, has_next_task=True):
        expected_tasks = []

        if has_next_task:
            task_spec = conductor.spec.tasks.get_task(task_id)
            expected_tasks = [self.format_task_item(task_id, route, ctx, task_spec)]

        self.assert_task_list(conductor, conductor.get_next_tasks(), expected_tasks)

    def assert_next_tasks(self, conductor, task_ids=None, ctxs=None, routes=None):
        expected_tasks = []

        for task_id, ctx, route in zip(task_ids, ctxs, routes):
            task_spec = conductor.spec.tasks.get_task(task_id)
            expected_tasks.append(self.format_task_item(task_id, route, ctx, task_spec))

        self.assert_task_list(conductor, conductor.get_next_tasks(), expected_tasks)

    def assert_conducting_sequences(
        self,
        wf_name,
        expected_task_seq,
        expected_routes=None,
        inputs=None,
        mock_statuses=None,
        mock_results=None,
        expected_workflow_status=None,
        expected_output=None,
        expected_term_tasks=None,
    ):

        if not expected_routes:
            expected_routes = [[]]

        if inputs is None:
            inputs = {}

        wf_def = self.get_wf_def(wf_name)
        wf_spec = self.spec_module.instantiate(wf_def)
        conductor = conducting.WorkflowConductor(wf_spec, inputs=inputs)
        conductor.request_workflow_status(statuses.RUNNING)

        run_q = queue.Queue()
        status_q = queue.Queue()
        result_q = queue.Queue()

        if mock_statuses:
            for item in mock_statuses:
                status_q.put(item)

        if mock_results:
            for item in mock_results:
                result_q.put(item)

        # Get start tasks and being conducting workflow.
        for task in conductor.get_next_tasks():
            run_q.put(task)

        # Serialize workflow conductor to mock async execution.
        wf_conducting_state = conductor.serialize()

        # Process until there are not more tasks in queue.
        while not run_q.empty():
            # Deserialize workflow conductor to mock async execution.
            conductor = conducting.WorkflowConductor.deserialize(wf_conducting_state)

            # Process all the tasks in the run queue.
            while not run_q.empty():
                current_task = run_q.get()
                current_task_id = current_task["id"]
                current_task_route = current_task["route"]

                # Set task status to running.
                ac_ex_event = events.ActionExecutionEvent(statuses.RUNNING)
                conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)

                # Mock completion of the task.
                status = status_q.get() if not status_q.empty() else statuses.SUCCEEDED
                result = result_q.get() if not result_q.empty() else None

                ac_ex_event = events.ActionExecutionEvent(status, result=result)
                conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)

            # Identify the next set of tasks.
            for next_task in conductor.get_next_tasks():
                run_q.put(next_task)

            # Serialize workflow execution graph to mock async execution.
            wf_conducting_state = conductor.serialize()

        actual_task_seq = [
            (entry["id"], entry["route"]) for entry in conductor.workflow_state.sequence
        ]

        expected_task_seq = [
            task_seq if isinstance(task_seq, tuple) else (task_seq, 0)
            for task_seq in expected_task_seq
        ]

        self.assertListEqual(actual_task_seq, expected_task_seq)
        self.assertListEqual(conductor.workflow_state.routes, expected_routes)

        if conductor.get_workflow_status() in statuses.COMPLETED_STATUSES:
            conductor.render_workflow_output()

        if expected_workflow_status is None:
            expected_workflow_status = statuses.SUCCEEDED

        self.assertEqual(conductor.get_workflow_status(), expected_workflow_status)

        if expected_output is not None:
            self.assertDictEqual(conductor.get_workflow_output(), expected_output)

        if expected_term_tasks:
            expected_term_tasks = [
                (task, 0) if not isinstance(task, tuple) else task for task in expected_term_tasks
            ]

            term_tasks = conductor.workflow_state.get_terminal_tasks()
            actual_term_tasks = [(t["id"], t["route"]) for i, t in term_tasks]
            expected_term_tasks = sorted(expected_term_tasks, key=lambda x: x[0])
            actual_term_tasks = sorted(actual_term_tasks, key=lambda x: x[0])
            self.assertListEqual(actual_term_tasks, expected_term_tasks)

        return conductor

    def assert_workflow_status(self, wf_name, mock_flow, expected_wf_statuses, conductor=None):
        if not conductor:
            wf_def = self.get_wf_def(wf_name)
            wf_spec = self.spec_module.instantiate(wf_def)
            conductor = conducting.WorkflowConductor(wf_spec)
            conductor.request_workflow_status(statuses.RUNNING)

        for _entry, expected_wf_status in zip(mock_flow, expected_wf_statuses):
            task_id = _entry["id"]
            task_status = _entry["status"]

            self.forward_task_statuses(conductor, task_id, [task_status])

            err_ctx = (
                'Workflow status "%s" is not the expected status "%s". '
                'Updated task "%s" with status "%s".'
                % (conductor.get_workflow_status(), expected_wf_status, task_id, task_status)
            )

            self.assertEqual(conductor.get_workflow_status(), expected_wf_status, err_ctx)

        return conductor


class WorkflowConductorWithItemsTest(WorkflowConductorTest):
    def assert_task_items(
        self,
        conductor,
        task_id,
        task_route,
        task_ctx,
        items,
        action_specs,
        mock_ac_ex_statuses,
        expected_task_statuses,
        expected_workflow_statuses,
        concurrency=None,
        mock_ac_ex_results=None,
    ):

        # Set up test cases.
        tests = list(zip(mock_ac_ex_statuses, expected_task_statuses, expected_workflow_statuses))
        tk_ex_result = [None] * len(items)

        if mock_ac_ex_results is None:
            mock_ac_ex_results = items

        # Verify the first set of action executions.
        expected_task = self.format_task_item(
            task_id,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_id),
            actions=action_specs[0:concurrency],
            items_count=len(items),
            items_concurrency=concurrency,
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # If items is an empty list, then mark the task as running.
        if len(items) == 0:
            ac_ex_event = events.ActionExecutionEvent(statuses.RUNNING)
            conductor.update_task_state(task_id, task_route, ac_ex_event)
        else:
            # Mark the action executions for the first set of items as running.
            for i in range(0, min(len(tests), concurrency or len(items))):
                ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.RUNNING)
                conductor.update_task_state(task_id, task_route, ac_ex_event)

        # Ensure the actions listed is accurate when getting next tasks again.
        expected_tasks = []
        capacity = len(items)
        next_item_id = len(tests)
        next_action_specs = []

        if concurrency is not None and concurrency > 0:
            items_running = min(len(tests), concurrency or len(items))
            capacity = concurrency - items_running

        if capacity > 0 and next_item_id < len(items):
            next_action_specs = action_specs[next_item_id : next_item_id + capacity]

        if next_action_specs or len(items) == 0:
            expected_task = self.format_task_item(
                task_id,
                task_route,
                task_ctx,
                conductor.spec.tasks.get_task(task_id),
                actions=next_action_specs,
                items_count=len(items),
                items_concurrency=concurrency,
            )

            expected_tasks = [expected_task]

        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # If items is an empty list, complete the task.
        if len(items) == 0:
            ac_ex_event = events.ActionExecutionEvent(statuses.SUCCEEDED)
            conductor.update_task_state(task_id, task_route, ac_ex_event)

        # Mock the action execution for each item.
        for item_id in range(0, len(tests)):
            ac_ex_result = mock_ac_ex_results[item_id]
            tk_ex_result[item_id] = ac_ex_result
            ac_ex_status = tests[item_id][0]

            ac_ex_event = events.TaskItemActionExecutionEvent(
                item_id, ac_ex_status, ac_ex_result, accumulated_result=tk_ex_result
            )

            conductor.update_task_state(task_id, task_route, ac_ex_event)

            expected_task_status = tests[item_id][1]
            actual_task_status = conductor.get_task_state_entry(task_id, task_route)["status"]

            error_message = 'Task execution status "%s" does not match "%s" for item %s.' % (
                actual_task_status,
                expected_task_status,
                item_id,
            )

            self.assertEqual(actual_task_status, expected_task_status, error_message)

            expected_workflow_status = tests[item_id][2]
            actual_workflow_status = conductor.get_workflow_status()

            error_message = (
                'Workflow execution status "%s" does not match "%s" after item %s update.'
                % (actual_workflow_status, expected_workflow_status, item_id)
            )

            self.assertEqual(actual_workflow_status, expected_workflow_status, error_message)

            # Process next set of action executions only if there are more test cases.
            if i >= len(tests) - 2 or concurrency is None:
                continue

            next_item_id = item_id + concurrency
            expected_tasks = []

            if next_item_id < len(items):
                expected_task = self.format_task_item(
                    task_id,
                    task_route,
                    task_ctx,
                    conductor.spec.tasks.get_task(task_id),
                    actions=action_specs[next_item_id : next_item_id + 1],
                    items_count=len(items),
                    items_concurrency=concurrency,
                )

                expected_tasks = [expected_task]

            actual_tasks = conductor.get_next_tasks()
            self.assert_task_list(conductor, actual_tasks, expected_tasks)

            for task in actual_tasks:
                task_id = task["id"]

                for action in task["actions"]:
                    item_id = action["item_id"]
                    ac_ex_event = events.TaskItemActionExecutionEvent(item_id, statuses.RUNNING)
                    conductor.update_task_state(task_id, task_route, ac_ex_event)


class WorkflowConductorRerunTest(WorkflowConductorTest):
    def assert_rerun_failed_tasks(
        self,
        conductor,
        expected_task_seq,
        expected_routes=None,
        rerun_tasks=None,
        mock_statuses=None,
        mock_results=None,
        expected_workflow_status=None,
        expected_output=None,
        expected_term_tasks=None,
    ):
        # Setup the test.
        run_q = queue.Queue()
        status_q = queue.Queue()
        result_q = queue.Queue()

        if not expected_routes:
            expected_routes = [[]]

        if mock_statuses:
            for item in mock_statuses:
                status_q.put(item)

        if mock_results:
            for item in mock_results:
                result_q.put(item)

        # Request workflow rerun and assert workflow status is running.
        conductor.request_workflow_rerun(task_requests=rerun_tasks)
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)

        # Get next tasks and resume conducting workflow.
        for task in conductor.get_next_tasks():
            run_q.put(task)

        # Serialize workflow conductor to mock async execution.
        wf_conducting_state = conductor.serialize()

        # Process until there are not more tasks in queue.
        while not run_q.empty():
            # Deserialize workflow conductor to mock async execution.
            conductor = conducting.WorkflowConductor.deserialize(wf_conducting_state)

            # Process all the tasks in the run queue.
            while not run_q.empty():
                current_task = run_q.get()
                current_task_id = current_task["id"]
                current_task_route = current_task["route"]

                # Set task status to running.
                ac_ex_event = events.ActionExecutionEvent(statuses.RUNNING)
                conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)

                # Mock completion of the task.
                status = status_q.get() if not status_q.empty() else statuses.SUCCEEDED
                result = result_q.get() if not result_q.empty() else None

                ac_ex_event = events.ActionExecutionEvent(status, result=result)
                conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)

            # Identify the next set of tasks.
            for next_task in conductor.get_next_tasks():
                run_q.put(next_task)

            # Serialize workflow execution graph to mock async execution.
            wf_conducting_state = conductor.serialize()

        actual_task_seq = [
            (entry["id"], entry["route"]) for entry in conductor.workflow_state.sequence
        ]

        expected_task_seq = [
            task_seq if isinstance(task_seq, tuple) else (task_seq, 0)
            for task_seq in expected_task_seq
        ]

        self.assertListEqual(actual_task_seq, expected_task_seq)
        self.assertListEqual(conductor.workflow_state.routes, expected_routes)

        if conductor.get_workflow_status() in statuses.COMPLETED_STATUSES:
            conductor.render_workflow_output()

        if expected_workflow_status is None:
            expected_workflow_status = statuses.SUCCEEDED

        self.assertEqual(conductor.get_workflow_status(), expected_workflow_status)

        if expected_output is not None:
            self.assertDictEqual(conductor.get_workflow_output(), expected_output)

        if expected_term_tasks:
            expected_term_tasks = [
                (task, 0) if not isinstance(task, tuple) else task for task in expected_term_tasks
            ]

            term_tasks = conductor.workflow_state.get_terminal_tasks()
            actual_term_tasks = [(t["id"], t["route"]) for i, t in term_tasks]
            expected_term_tasks = sorted(expected_term_tasks, key=lambda x: x[0])
            actual_term_tasks = sorted(actual_term_tasks, key=lambda x: x[0])
            self.assertListEqual(actual_term_tasks, expected_term_tasks)

        return conductor
